/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef EXTERN_H
#define EXTERN_H

/*
 * All files sourcing extern.h will need this for ohash and pretty much
 * nothing else.
 * Thus, it's a little cleaner to include it here (for now?).
 */
#include <stdint.h>
#include <stddef.h>
#ifdef HAVE_OHASH
# include <ohash.h>
#else
# include "compat_ohash.h"
#endif

enum	merror {
	MERROR_OK = 0,
#define	MERROR_WARNING	 MERROR_QUOTE /* just warnings */
	MERROR_QUOTE,
	MERROR_ARGSMANY,
	MERROR_ESCAPE,
	MERROR_EEOF,
	MERROR_USEGZIP,
	MERROR_TXT_NO_CTX,
	MERROR_BROKEN_SCOPE,
	MERROR_BAD_SCOPE,
#define	MERROR_ERROR	 MERROR_MACRO /* errors */
	MERROR_MACRO,
	MERROR_BADESCAPE,
	MERROR_WCS,
	MERROR_EXITSCOPE,
	MERROR_ARGSFEW,
	MERROR_NAME,
	MERROR_NAMEDEF,
	MERROR_STRINGDEF,
	MERROR_STACK,
	MERROR_MAN_IN_MDOC,
	MERROR_MDOC_IN_MAN,
#define	MERROR_FATAL	 MERROR_OPEN /* system call errors */
	MERROR_OPEN,
	MERROR_STAT,
	MERROR_ZLIB,
	MERROR_MMAP,
	MERROR_READ,
	MERROR__MAX
};

/*
 * This consists of all macros we'll parse.
 * This isn't to be confused with all macros we'll export to the AST:
 * those can be found in mandoc.h as enum melem.
 */
enum	mdtype {
	MDTYPE__TEXT = 0,
	MDTYPE_Ac,
	MDTYPE_Ad,
	MDTYPE_An,
	MDTYPE_Ao,
	MDTYPE_Ap,
	MDTYPE_Aq,
	MDTYPE_Ar,
	MDTYPE_At,
	MDTYPE_Bc,
	MDTYPE_Bd,
	MDTYPE_Bf,
	MDTYPE_Bk,
	MDTYPE_Bl,
	MDTYPE_Bo,
	MDTYPE_Bq,
	MDTYPE_Brc,
	MDTYPE_Bro,
	MDTYPE_Brq,
	MDTYPE_Bsx,
	MDTYPE_Bt,
	MDTYPE_Bx,
	MDTYPE_Cd,
	MDTYPE_Cm,
	MDTYPE_D1,
	MDTYPE_Db,
	MDTYPE_Dc,
	MDTYPE_Dd,
	MDTYPE_Dl,
	MDTYPE_Do,
	MDTYPE_Dq,
	MDTYPE_Dt,
	MDTYPE_Dv,
	MDTYPE_Dx,
	MDTYPE_Ec,
	MDTYPE_Ed,
	MDTYPE_Ef,
	MDTYPE_Ek,
	MDTYPE_El,
	MDTYPE_Em,
	MDTYPE_En,
	MDTYPE_Eo,
	MDTYPE_Er,
	MDTYPE_Es,
	MDTYPE_Ev,
	MDTYPE_Ex,
	MDTYPE_Fa,
	MDTYPE_Fc,
	MDTYPE_Fd,
	MDTYPE_Fl,
	MDTYPE_Fn,
	MDTYPE_Fo,
	MDTYPE_Fr,
	MDTYPE_Ft,
	MDTYPE_Fx,
	MDTYPE_Hf,
	MDTYPE_Ic,
	MDTYPE_In,
	MDTYPE_It,
	MDTYPE_Lb,
	MDTYPE_Li,
	MDTYPE_Lk,
	MDTYPE_Lp,
	MDTYPE_Ms,
	MDTYPE_Mt,
	MDTYPE_Nd,
	MDTYPE_Nm,
	MDTYPE_No,
	MDTYPE_Ns,
	MDTYPE_Nx,
	MDTYPE_Oc,
	MDTYPE_Oo,
	MDTYPE_Op,
	MDTYPE_Os,
	MDTYPE_Ot,
	MDTYPE_Ox,
	MDTYPE_Pa,
	MDTYPE_Pc,
	MDTYPE_Pf,
	MDTYPE_Po,
	MDTYPE_Pp,
	MDTYPE_Pq,
	MDTYPE_Qc,
	MDTYPE_Ql,
	MDTYPE_Qo,
	MDTYPE_Qq,
	MDTYPE_Re,
	MDTYPE_Rs,
	MDTYPE_Rv,
	MDTYPE_Sc,
	MDTYPE_Sh,
	MDTYPE_Sm,
	MDTYPE_So,
	MDTYPE_Sq,
	MDTYPE_Ss,
	MDTYPE_St,
	MDTYPE_Sx,
	MDTYPE_Sy,
	MDTYPE_TH,
	MDTYPE_Ta,
	MDTYPE_Tn,
	MDTYPE_Ud,
	MDTYPE_Ux,
	MDTYPE_Va,
	MDTYPE_Vt,
	MDTYPE_Xc,
	MDTYPE_Xo,
	MDTYPE_Xr,
	MDTYPE__A,
	MDTYPE__B,
	MDTYPE__C,
	MDTYPE__D,
	MDTYPE__I,
	MDTYPE__J,
	MDTYPE__N,
	MDTYPE__O,
	MDTYPE__P,
	MDTYPE__Q,
	MDTYPE__R,
	MDTYPE__T,
	MDTYPE__U,
	MDTYPE__V,
	MDTYPE_br,
	MDTYPE_c2,
	MDTYPE_cc,
	MDTYPE_de,
	MDTYPE_de1,
	MDTYPE_ds,
	MDTYPE_ds1,
	MDTYPE_ec,
	MDTYPE_ecr,
	MDTYPE_ecs,
	MDTYPE_el,
	MDTYPE_ie,
	MDTYPE_if,
	MDTYPE_ig,
	MDTYPE_so,
	MDTYPE_sp,
	MDTYPE_while,
	MDTYPE__USERDEF,
	MDTYPE__NONE,
	MDTYPE__MAX
};

struct	mparse; /* parse sequence (file + `so'd files) */
struct	mfile; /* single file in parse sequence */
struct	mline; /* line within a file */
struct	mdmacro; /* defined macros in parse sequence */

typedef	void (*mmacro)(struct mline **);
typedef	void (*mlparse)(struct mline *, wchar_t *, size_t);

struct	mdmacro {
	enum mdtype	 type; /* macro type */
	mmacro		 macro; /* processing function */
	mlparse		 parser; /* argument parsing function */
	struct mline	*first; /* first line of contents */
	struct mline	*last; /* last line (inclusive) */
	wchar_t		 name[]; /* name (e.g., `cc') */
};

enum	mmode {
	MMODE_NROFF = 0, /* "text" mode */
	MMODE_TROFF /* "graphical" mode */
};

enum	minput {
	MINPUT_UNSET = 0, /* not yet man or mdoc input */
	MINPUT_MDOC, /* is absolutely an mdoc file */
	MINPUT_MAN /* is absolutely a man file */
};

struct	mstacked {
	struct mline	*line; /* line weak scope was parsed */
	enum mdtype	 type; /* type of weak scope */
	struct mnode	*n; /* node at col/line/etc. */
};

struct	mparse {
	wchar_t		 esc; /* escape character */
	wchar_t		 escs; /* saved escape character */
	wchar_t		 ctr; /* control character */
	wchar_t		 lctr; /* line-control character */
	enum mlevel	 level; /* current max warned level */
	enum mlevel	 wlevel; /* warning level */
	struct mfile	*file; /* list of parsed files */
	mmsg		 msg; /* errors/warnings */
	struct ohash	 macros; /* all defined macros */
	struct mline	*dynlines; /* list of dynamic lines */
	int		 cond; /* last conditional (Boolean) */
	enum mmode	 mode; /* whether troff/nroff mode */
	enum minput	 input; /* type of input document */
	struct mnode	*node; /* last parsed node */
	struct mnode	*root; /* first parsed node */
#define	MDOC_STACKSZ	 8
	/*
	 * In mdoc(7), we can have "weakly nested" scopes, i.e., scopes
	 * that can be broken (e.g., Ao).
	 * Here we maintain a stack of these to manage the breakage.
	 */
	struct mstacked	 wstack[MDOC_STACKSZ];
	size_t		 wstackpos;
	/*
	 * Like the "wstack", we also maintain a stack of line macros
	 * (e.g., Aq).
	 * This is because a weakly-scoped macro can open within such a
	 * macro and prolong its scope, although we want to record its
	 * ending on the line.
	 */
	struct mstacked	 lstack[MDOC_STACKSZ];
	size_t		 lstackpos;
};

struct	mline {
	const wchar_t	 *line; /* original text of line */
	wchar_t		 *copy; /* copy-mode line */
	wchar_t		 *wline; /* working form of line */
	wchar_t		  ctr; /* macro control character or \0 */
	enum mdtype	  macro; /* macro or MDTYPE__MAX */
	const wchar_t	**argv; /* parsed args */
	size_t		 *col; /* starting column of parsed args */
	size_t		  argc; /* number of parsed args */
	size_t		  maxarg; /* max size of parsed arg buffer */
	size_t		  lineno; /* current line number */
	struct mfile	 *file; /* line's file */
#define	MLINE_STACKSZ	  4
	struct mline	 *next[MLINE_STACKSZ]; /* next position */
	size_t		  stackpos; /* allocated recursion depth */
	struct mline	 *dynnext; /* list of dynamic lines */
	int		  clean; /* parsed; don't parse again */
};

struct	mfile {
	struct mparse	*parse; /* parse sequence */
	struct mline	*lines; /* lines in file */
	size_t		 sz; /* number of lines */
	char		*fname; /* filename */
	wchar_t		*fbuf; /* buffer of contents */
	struct mfile	*next; /* next parsed file or NULL */
};

__BEGIN_DECLS

struct mdmacro	*mdmacro_calloc(const wchar_t *name);
void		 mdmacro_clear(struct mparse *mp);
struct mdmacro	*mdmacro_find(struct mparse *mp, 
			const wchar_t *name);
struct mdmacro	*mdmacro_finds(struct mparse *mp, 
			const wchar_t *start, const wchar_t *end);
void		 mdmacro_init(struct mparse *mp);
unsigned int	 mdmacro_lookup(struct mparse *mp, 
			const wchar_t *name);
unsigned int	 mdmacro_lookups(struct mparse *mp, 
			const wchar_t *start, const wchar_t *end);
void		 mdmacro_zero(struct mdmacro *md);

void	 	 merr(struct mparse *mp, enum merror e, 
			const char *f, size_t ln, size_t col);
void	 	 merrx(struct mparse *mp, enum merror e, 
			const char *f, size_t ln, size_t col, 
			const wchar_t *fmt, ...);
void		 mlerrx(struct mline *l, enum merror e,
			size_t col, const wchar_t *fmt, ...);

void 	 	 mfile_free(struct mfile *mf);
int	 	 mfile_next(struct mparse *mp, 
			const char *f, struct mline *l);

struct mline	*mline_dynalloc(struct mline *l, const wchar_t *cp);
void		 mline_cc(struct mline *l, wchar_t *cp, size_t i);
void		 mline_de(struct mline *l, wchar_t *cp, size_t i);
void		 mline_ds(struct mline *l, wchar_t *cp, size_t i);
void		 mline_ecr(struct mline *l, wchar_t *cp, size_t i);
void		 mline_el(struct mline *l, wchar_t *cp, size_t i);
void	 	 mline_free(struct mline *l);
void		 mline_if(struct mline *l, wchar_t *cp, size_t i);
void	 	 mline_parse(struct mline *ml);
struct mline	*mline_peek(struct mline *l);
struct mline	*mline_popnext(struct mline *l);
int		 mline_pushnext(struct mline *l, struct mline *next);

void		 mmacro_char(struct mline **lp);
void		 mmacro_de(struct mline **lp);
void		 mmacro_ds(struct mline **lp);
void		 mmacro_ecr(struct mline **lp);
void		 mmacro_el(struct mline **lp);
void		 mmacro_next(struct mline **lp);
void		 mmacro_ie(struct mline **lp);
void		 mmacro_if(struct mline **lp);
void		 mmacro_ig(struct mline **lp);
void		 mmacro_man(struct mline **lp);
void		 mmacro_mdoc(struct mline **lp);
void		 mmacro_so(struct mline **lp);
void		 mmacro_while(struct mline **lp);

void		 mnode_alloc_node(struct mline *l, size_t pos, 
			enum mdtype hint);
void		 mnode_alloc_node_line(struct mline *l, 
			enum mdtype hint);
void		 mnode_alloc_text(struct mline *l, size_t pos);
void		 mnode_alloc_text_all(struct mline *l);
void		 mnode_free_all(struct mnode *);

size_t	 	 munescape(struct mline *l, wchar_t **dst);

__END_DECLS

#endif /*!EXTERN_H*/
