/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <getopt.h>
#include <locale.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include "mandoc.h"

#define	UNCONST(a) ((void *)(uintptr_t)(const void *)(a))

static	void	 mandoc_msg(enum mlevel e, const wchar_t *msg);
static	int	 woptions(enum mlevel *warnp, int *stop, char *arg);

static void
mandoc_msg(enum mlevel l, const wchar_t *msg)
{

	fwprintf(stderr, L"%ls\n", msg);
}

static void
print_node(struct mnode *n, size_t indent)
{
	size_t		 i;

	if (NULL == n)
		return;

	for ( ; NULL != n; n = n->next) {
		for (i = 0; i < indent; i++)
			putwchar(L'\t');
		wprintf(L"node (");
		if (NULL != n->string) {
			assert(MELEM__TEXT == n->elem);
			wprintf(L"`%ls'", n->string);
		} else
			wprintf(L"%d", n->elem);
		if (NULL != n->broken) 
			wprintf(L", broken");
		wprintf(L")\n");
		print_node(n->first, indent + 1);
	}
}

int
main(int argc, char *argv[])
{
	int		 ch, i, stop;
	struct mparse	*mp;
	const char	*progname;
	enum mlevel	 warn, e, er = MLEVEL_OK;

	setlocale(LC_CTYPE, "");

	progname = strrchr(argv[0], '/');
	if (progname == NULL)
		progname = argv[0];
	else
		++progname;

	warn = MLEVEL_FATAL;
	stop = 0;

	while (-1 != (ch = getopt(argc, argv, "W:")))
		switch (ch) {
		case ('W'):
			if ( ! woptions(&warn, &stop, optarg))
				goto usage;
			break;
		default:
			goto usage;
		}

	argc -= optind;
	argv += optind;

	mp = mparse_alloc();

	for (i = 0; i < argc; i++) {
		mparse_init(mp, warn, mandoc_msg);
		e = mparse_parse(mp, argv[i]);
		/*print_node(mparse_result(mp), 0);*/
		mparse_clear(mp);
		if (e > er)
			er = e;
		if (er >= warn && stop)
			break;
	}

	mparse_free(mp);
	return(er);
usage:
	fprintf(stderr, "usage: %s [-Wlevel] file...\n", progname);
	return(MLEVEL_BADARG);
}

static int
woptions(enum mlevel *warnp, int *stop, char *arg)
{
	char		*v, *o;
	const char	*toks[7]; 

	toks[0] = "all";
	toks[1] = "debug";
	toks[2] = "warning";
	toks[3] = "error";
	toks[4] = "fatal";
	toks[5] = "stop";
	toks[6] = NULL;

	while (*arg) {
		o = arg;
		switch (getsubopt(&arg, UNCONST(toks), &v)) {
		case (1):
			*warnp = MLEVEL_OK;
			break;
		case (0):
			/* FALLTHROUGH */
		case (2):
			*warnp = MLEVEL_WARNING;
			break;
		case (3):
			*warnp = MLEVEL_ERROR;
			break;
		case (4):
			*warnp = MLEVEL_FATAL;
			break;
		case (5):
			*stop = 1;
			break;
		default:
			fprintf(stderr, "-W%s: Bad argument\n", o);
			return(0);
		}
	}

	return(1);
}

