/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/param.h>

#include <assert.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

struct	man {
	mmacro		 macro;
};

static	void	man_TH(struct mline **lp);

static	const struct man mans[MDTYPE__MAX] = {
	{ NULL }, /* MDTYPE__TEXT */
	{ NULL }, /* MDTYPE_Ac */
	{ NULL }, /* MDTYPE_Ad */
	{ NULL }, /* MDTYPE_An */
	{ NULL }, /* MDTYPE_Ao */
	{ NULL }, /* MDTYPE_Ap */
	{ NULL }, /* MDTYPE_Aq */
	{ NULL }, /* MDTYPE_Ar */
	{ NULL }, /* MDTYPE_At */
	{ NULL }, /* MDTYPE_Bc */
	{ NULL }, /* MDTYPE_Bd */
	{ NULL }, /* MDTYPE_Bf */
	{ NULL }, /* MDTYPE_Bk */
	{ NULL }, /* MDTYPE_Bl */
	{ NULL }, /* MDTYPE_Bo */
	{ NULL }, /* MDTYPE_Bq */
	{ NULL }, /* MDTYPE_Brc */
	{ NULL }, /* MDTYPE_Bro */
	{ NULL }, /* MDTYPE_Brq */
	{ NULL }, /* MDTYPE_Bsx */
	{ NULL }, /* MDTYPE_Bt */
	{ NULL }, /* MDTYPE_Bx */
	{ NULL }, /* MDTYPE_Cd */
	{ NULL }, /* MDTYPE_Cm */
	{ NULL }, /* MDTYPE_D1 */
	{ NULL }, /* MDTYPE_Db */
	{ NULL }, /* MDTYPE_Dc */
	{ NULL }, /* MDTYPE_Dd */
	{ NULL }, /* MDTYPE_Dl */
	{ NULL }, /* MDTYPE_Do */
	{ NULL }, /* MDTYPE_Dq */
	{ NULL }, /* MDTYPE_Dt */
	{ NULL }, /* MDTYPE_Dv */
	{ NULL }, /* MDTYPE_Dx */
	{ NULL }, /* MDTYPE_Ec */
	{ NULL }, /* MDTYPE_Ed */
	{ NULL }, /* MDTYPE_Ef */
	{ NULL }, /* MDTYPE_Ek */
	{ NULL }, /* MDTYPE_El */
	{ NULL }, /* MDTYPE_Em */
	{ NULL }, /* MDTYPE_En */
	{ NULL }, /* MDTYPE_Eo */
	{ NULL }, /* MDTYPE_Er */
	{ NULL }, /* MDTYPE_Es */
	{ NULL }, /* MDTYPE_Ev */
	{ NULL }, /* MDTYPE_Ex */
	{ NULL }, /* MDTYPE_Fa */
	{ NULL }, /* MDTYPE_Fc */
	{ NULL }, /* MDTYPE_Fd */
	{ NULL }, /* MDTYPE_Fl */
	{ NULL }, /* MDTYPE_Fn */
	{ NULL }, /* MDTYPE_Fo */
	{ NULL }, /* MDTYPE_Fr */
	{ NULL }, /* MDTYPE_Ft */
	{ NULL }, /* MDTYPE_Fx */
	{ NULL }, /* MDTYPE_Hf */
	{ NULL }, /* MDTYPE_Ic */
	{ NULL }, /* MDTYPE_In */
	{ NULL }, /* MDTYPE_It */
	{ NULL }, /* MDTYPE_Lb */
	{ NULL }, /* MDTYPE_Li */
	{ NULL }, /* MDTYPE_Lk */
	{ NULL }, /* MDTYPE_Lp */
	{ NULL }, /* MDTYPE_Ms */
	{ NULL }, /* MDTYPE_Mt */
	{ NULL }, /* MDTYPE_Nd */
	{ NULL }, /* MDTYPE_Nm */
	{ NULL }, /* MDTYPE_No */
	{ NULL }, /* MDTYPE_Ns */
	{ NULL }, /* MDTYPE_Nx */
	{ NULL }, /* MDTYPE_Oc */
	{ NULL }, /* MDTYPE_Oo */
	{ NULL }, /* MDTYPE_Op */
	{ NULL }, /* MDTYPE_Os */
	{ NULL }, /* MDTYPE_Ot */
	{ NULL }, /* MDTYPE_Ox */
	{ NULL }, /* MDTYPE_Pa */
	{ NULL }, /* MDTYPE_Pc */
	{ NULL }, /* MDTYPE_Pf */
	{ NULL }, /* MDTYPE_Po */
	{ NULL }, /* MDTYPE_Pp */
	{ NULL }, /* MDTYPE_Pq */
	{ NULL }, /* MDTYPE_Qc */
	{ NULL }, /* MDTYPE_Ql */
	{ NULL }, /* MDTYPE_Qo */
	{ NULL }, /* MDTYPE_Qq */
	{ NULL }, /* MDTYPE_Re */
	{ NULL }, /* MDTYPE_Rs */
	{ NULL }, /* MDTYPE_Rv */
	{ NULL }, /* MDTYPE_Sc */
	{ NULL }, /* MDTYPE_Sh */
	{ NULL }, /* MDTYPE_Sm */
	{ NULL }, /* MDTYPE_So */
	{ NULL }, /* MDTYPE_Sq */
	{ NULL }, /* MDTYPE_Ss */
	{ NULL }, /* MDTYPE_St */
	{ NULL }, /* MDTYPE_Sx */
	{ NULL }, /* MDTYPE_Sy */
	{ man_TH }, /* MDTYPE_TH */
	{ NULL }, /* MDTYPE_Ta */
	{ NULL }, /* MDTYPE_Tn */
	{ NULL }, /* MDTYPE_Ud */
	{ NULL }, /* MDTYPE_Ux */
	{ NULL }, /* MDTYPE_Va */
	{ NULL }, /* MDTYPE_Vt */
	{ NULL }, /* MDTYPE_Xc */
	{ NULL }, /* MDTYPE_Xo */
	{ NULL }, /* MDTYPE_Xr */
	{ NULL }, /* MDTYPE__A */
	{ NULL }, /* MDTYPE__B */
	{ NULL }, /* MDTYPE__C */
	{ NULL }, /* MDTYPE__D */
	{ NULL }, /* MDTYPE__I */
	{ NULL }, /* MDTYPE__J */
	{ NULL }, /* MDTYPE__N */
	{ NULL }, /* MDTYPE__O */
	{ NULL }, /* MDTYPE__P */
	{ NULL }, /* MDTYPE__Q */
	{ NULL }, /* MDTYPE__R */
	{ NULL }, /* MDTYPE__T */
	{ NULL }, /* MDTYPE__U */
	{ NULL }, /* MDTYPE__V */
	{ NULL }, /* MDTYPE_br */
	{ NULL }, /* MDTYPE_c2 */
	{ NULL }, /* MDTYPE_cc */
	{ NULL }, /* MDTYPE_de */
	{ NULL }, /* MDTYPE_de1 */
	{ NULL }, /* MDTYPE_ds */
	{ NULL }, /* MDTYPE_ds1 */
	{ NULL }, /* MDTYPE_ec */
	{ NULL }, /* MDTYPE_ecr */
	{ NULL }, /* MDTYPE_ecs */
	{ NULL }, /* MDTYPE_el */
	{ NULL }, /* MDTYPE_ie */
	{ NULL }, /* MDTYPE_if */
	{ NULL }, /* MDTYPE_ig */
	{ NULL }, /* MDTYPE_so */
	{ NULL }, /* MDTYPE_sp */
	{ NULL }, /* MDTYPE_while */
	{ NULL }, /* MDTYPE__USERDEF */
	{ NULL }, /* MDTYPE__NONE */
};

/*
 * Route all man(7) macros through this function.
 * We do this instead of the roff(7) approach to catch whether we're
 * mixing macro contexts.
 */
void
mmacro_man(struct mline **lp)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mline	*l = *lp;

	if (MINPUT_MDOC == mp->input) {
		mlerrx(l, MERROR_MDOC_IN_MAN,
			l->col[0], L"%ls", l->argv[0]);
		*lp = mline_popnext(*lp);
		return;
	} else if (MINPUT_UNSET == mp->input)
		mp->input = MINPUT_MAN;

	assert(NULL != mans[l->macro].macro);
	mans[l->macro].macro(lp);
}

static void
man_TH(struct mline **lp)
{

	*lp = mline_popnext(*lp);
}
