/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef MANDOC_H
#define MANDOC_H

enum	mlevel {
	MLEVEL_OK = 0,
	MLEVEL_RESERVED,
	MLEVEL_WARNING, /* warnings: syntax, whitespace, etc. */
	MLEVEL_ERROR, /* input has been thrown away */
	MLEVEL_FATAL, /* input is borked */
	MLEVEL_BADARG, /* bad argument in invocation */
	MLEVEL_SYSTEM, /* system error */
	MLEVEL__MAX
};

enum	melem {
	MELEM__TEXT = 0,
	MELEM__ROOT,
	MELEM_Ac,
	MELEM_Ao,
	MELEM_Dt,
	MELEM_Oc,
	MELEM_Oo,
	MELEM_Op,
	MELEM_Sh,
	MELEM_Ss,
	MELEM__MAX
};

struct	mnode {
	struct mnode	*first;
	struct mnode	*parent;
	struct mnode	*next;
	struct mnode	*last;
	/*
	 * If a weakly-scoped block is broken, this will point to the
	 * offending part.
	 */
	struct mnode	*broken;
	enum melem	 elem;
	wchar_t		*string;
	size_t		 line;
	size_t		 pos;
};

typedef	void (*mmsg)(enum mlevel, const wchar_t *msg);

__BEGIN_DECLS

struct	mparse; 

void	*mandoc_calloc(size_t n, size_t sz);
void	*mandoc_malloc(size_t n);
void	*mandoc_realloc(void *p, size_t n);
char	*mandoc_strdup(const char *p);
wchar_t	*mandoc_wcsdup(const wchar_t *p);

struct mparse *mparse_alloc(void);
void	 mparse_clear(struct mparse *mp);
void	 mparse_free(struct mparse *mp);
void	 mparse_init(struct mparse *mp, enum mlevel l, mmsg msg);
enum mlevel mparse_parse(struct mparse *mp, const char *f);
struct mnode *mparse_result(struct mparse *mp);

__END_DECLS

#endif /*!MANDOC_H*/
