/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

void *
mandoc_malloc(size_t n)
{
	void		*p;

	if (NULL != (p = malloc(n))) 
		return(p);

	perror(NULL);
	exit(MLEVEL_SYSTEM);
}

void *
mandoc_realloc(void *p, size_t n)
{

	if (NULL != (p = realloc(p, n))) 
		return(p);

	perror(NULL);
	exit(MLEVEL_SYSTEM);
}

void *
mandoc_calloc(size_t n, size_t sz)
{
	void		*p;

	if (NULL != (p = calloc(n, sz))) 
		return(p);

	perror(NULL);
	exit(MLEVEL_SYSTEM);
}

char *
mandoc_strdup(const char *p)
{
	char		*cp;

	if (NULL != (cp = strdup(p))) 
		return(cp);

	perror(NULL);
	exit(MLEVEL_SYSTEM);
}

/*
 * Like wcsdup() but uses realloc() internally.
 * Use the copy-buffer if it exists (i.e., we've already been invoked in
 * a `de'), else use the real line.
 */
wchar_t *
mandoc_wcsdup(const wchar_t *cp)
{
	size_t		 sz;
	wchar_t		*rp;

	sz = (wcslen(cp) + 1) * sizeof(wchar_t);
	rp = mandoc_malloc(sz);
	memcpy(rp, cp, sz);
	return(rp);
}
