/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <stdlib.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

/*
 * Parse the "lp" line and, if a macro is found, branch into that
 * macro's processing.
 * Sets "lp" to NULL if an error has occured or end-of-file.
 * Otherwise, "lp" is set to the next line after full processing of the
 * input "lp" is finished.
 * Note that this can be many lines later!
 */
void
mmacro_next(struct mline **lp)
{
	struct mline	*l = *lp;
	struct mparse	*mp = (*lp)->file->parse;
	struct mdmacro	*md;

	mline_parse(l);
	l->clean = 0;
		
	switch (l->macro) {
	case (MDTYPE__TEXT):
		if (MINPUT_UNSET == mp->input)
			mlerrx(l, MERROR_TXT_NO_CTX, 
				l->col[0], NULL);
		mnode_alloc_text_all(l);
		*lp = mline_popnext(l);
		break;
	case (MDTYPE__NONE):
		*lp = mline_popnext(l);
		break;
	default:
		assert(l->argc > 0);
		assert(L'\0' != l->ctr);
		md = mdmacro_find
			((*lp)->file->parse, l->argv[0]);
		assert(NULL != md && NULL != md->macro);
		(*md->macro)(lp);
		break;
	} 
}
